% This file simulates the Batch Reactor Example (3.2.1) from Haseltine
% & Rawlings (2005) UKF is implemented to this example and it is shown that it
% fails when a poor prior is used.
data = struct();

% Helper functions.
function [x_sigma, w_sigma, v_sigma] = get_sigma(Pz, xhat, pars)
    % [x_sigma, w_sigma, v_sigma] = get_sigma(Pz, pars)
    %
    % Returns sigma points obtained from scaled sqrtm(Pz).
    narginchk(3, 3);

    % Make sure covariance is positive-definite.
    [Pzchol, okay] = chol(Pz);
    if okay ~= 0
        warning('Pz is not positive definite!');
    end

    % Select sigma points.
    sqrtPz = sqrtm((pars.Nz + pars.lambda)*Pz);
    zhat = [xhat; zeros(pars.Nw + pars.Nv, 1)];
    z_sigma = bsxfun(@plus, [zeros(pars.Nz, 1), sqrtPz, -sqrtPz], zhat);

    % Split into pieces.
    x_sigma = z_sigma(1:pars.Nx,:);
    w_sigma = z_sigma(pars.Nx + (1:pars.Nw),:);
    v_sigma = z_sigma(pars.Nx + pars.Nw + (1:pars.Nv),:);
end%function

function [x_sigma, weights] = ukf_scale(x_sigma, pars)
    % [x_sigma, weights] = ukf_scale(x_sigma, pars)
    %
    % Rescales the sigma points and adjusts weights so that each sigma point has
    % no negative components.
    %
    % x_sigma should be a matrix of column vectors with the first vector giving
    % the central point (which must always be feasible).
    x_center = x_sigma(:,1); % Central point.
    x_outer = x_sigma(:,2:end);

    % Choose theta to make each point feasible.
    dx = abs(bsxfun(@minus, x_outer, x_center));
    err = abs(min(x_outer, 0)); % Gets negative part.
    err(x_center < 0,:) = 0; % Can't fix indices where x_center is negative.
    dx(err == 0) = 1; % Avoid division by zero.
    theta = 1 - max(err./dx);

    % Rescale outer points.
    x_outer = bsxfun(@times, theta, x_outer) ...
              + bsxfun(@times, 1 - theta, x_center);
    x_sigma = [x_center, x_outer];

    % Calculate weights.
    rho = (2*pars.lambda - 1)/(pars.Nsigma - sum(theta));
    a = -rho/(2*(pars.Nz + pars.lambda));
    b = (1 + rho)/(2*(pars.Nz + pars.lambda));
    weights = [0, a*theta]' + b;
end%function

function data = doukf(Nsim, clipping)
    % data = doukf(Nsim, clipping)
    %
    % Simulate the UKF with or without clipping.
    pars = batchreactor(Nsim);
    Nx = pars.Nx;
    Ny = pars.Ny;
    Nw = pars.Nw;
    Nv = pars.Nv;
    pars.Nz = Nx + Nw + Nv; % Augmented state is z = [x; w; v].
    Nz = pars.Nz;

    % Choose simulation parameters.
    Nsim = pars.Nsim;
    t = pars.t;
    x = pars.xsim;
    y = pars.ysim;
    Px = pars.P;
    Qw = pars.Q;
    Rv = pars.R;

    % Choose UKF weights.
    Nsigma = 1 + 2*Nz;
    pars.Nsigma = Nsigma;
    pars.lambda = 1; % From Vachhani et. al. (2006). Guarantees Pz > 0.
    pars.weights = [2*pars.lambda; ones(2*Nz, 1)]/(2*(Nz + pars.lambda));
    W = diag(pars.weights);

    % Preallocate variables.
    xhat = NaN(Nx, Nsim + 1);
    yhat = NaN(Ny, Nsim + 1);
    e = NaN(Ny, Nsim + 1);

    xhat(:,1) = pars.xhat0; % Poor initial guess
    zhat(1:Nx,1) = pars.xhat0;

    Pz = blkdiag(Px, Qw, Rv); % Covariance for extended state.
    [x_sigma, w_sigma, v_sigma] = get_sigma(Pz, xhat(:,1), pars);

    for t = 1:(Nsim + 1)
        % Take measurement.
        y_sigma = NaN(Ny, Nsigma);
        for i = 1:Nsigma
            y_sigma(:,i) = full(pars.h(x_sigma(:,i))) + v_sigma(:,i);
        end
        yhat(:,t) = y_sigma*pars.weights;
        e(:,t) = y(:,t) - yhat(:,t);

        % Apply correction.
        Ex = bsxfun(@minus, x_sigma, xhat(:,t));
        Ey = bsxfun(@minus, y_sigma, yhat(:,t));

        Exy = Ex*W*Ey';
        Eyy = Ey*W*Ey';

        L = Exy/Eyy;

        Px = Px - L*Exy';
        xhat(:,t) = xhat(:,t) + L*e(:,t);

        % Stop if at end of simulation.
        if t > Nsim
            break
        end

        % Calculate new sigma points, and clip if necessary.
        Pz = blkdiag(Px, Qw, Rv);
        [x_sigma, w_sigma, v_sigma] = get_sigma(Pz, xhat(:,t), pars);
        if clipping
            [x_sigma, pars.weights] = ukf_scale(x_sigma, pars);
        end

        % Simulate sigma points and calculate new xhat and Px.
        for i = 1:Nsigma
            x_sigma(:,i) = full(pars.f(x_sigma(:,i), w_sigma(:,i)));
        end
        xhat(:,t + 1) = x_sigma*pars.weights;
        Ex = bsxfun(@minus, x_sigma, xhat(:,t + 1));
        Px = Ex*W*Ex';
    end

    % Make a plot.
    pars.reactorplot(x, xhat, y, yhat, pars.yclean, pars.Delta);
    set(gca(), 'ylim', [-1, 1.5]);

    % Form data table.
    data = [pars.t', x', xhat', y', yhat'];
end%function

% Simulate.
fprintf('Simulating without clipping.\n');
data.noclip = doukf(150, false());
fprintf('Simulating with clipping.\n');
data.yesclip = doukf(180, true());

